@@ -3,10 +3,11 @@
 # This file is released under the GNU GPL, version 3 or a later revision.
 # For further details see the COPYING file
 from collections import deque
+import contextlib
 import logging
 
-from notmuch import Database, NotmuchError, XapianError
-import notmuch
+from notmuch2 import Database, NotmuchError, XapianError
+import notmuch2
 
 from .errors import DatabaseError
 from .errors import DatabaseLockedError
@@ -25,10 +26,10 @@ class DBManager:
     classes.
     """
     _sort_orders = {
-        'oldest_first': notmuch.database.Query.SORT.OLDEST_FIRST,
-        'newest_first': notmuch.database.Query.SORT.NEWEST_FIRST,
-        'unsorted': notmuch.database.Query.SORT.UNSORTED,
-        'message_id': notmuch.database.Query.SORT.MESSAGE_ID,
+        'oldest_first': notmuch2.Database.SORT.OLDEST_FIRST,
+        'newest_first': notmuch2.Database.SORT.NEWEST_FIRST,
+        'unsorted': notmuch2.Database.SORT.UNSORTED,
+        'message_id': notmuch2.Database.SORT.MESSAGE_ID,
     }
     """constants representing sort orders"""
 
@@ -44,10 +45,20 @@ class DBManager:
         self.writequeue = deque([])
         self.processes = []
 
+    @property
+    def exclude_tags(self):
+        exclude_tags = settings.get('exclude_tags')
+        if exclude_tags is not None:
+            return exclude_tags
+
+        exclude_tags = settings.get_notmuch_setting('search', 'exclude_tags')
+        if exclude_tags:
+            return [t for t in exclude_tags.split(';') if t]
+
     def flush(self):
         """
         write out all queued write-commands in order, each one in a separate
-        :meth:`atomic <notmuch.Database.begin_atomic>` transaction.
+        :meth:`atomic <notmuch2.Database.atomic>` transaction.
 
         If this fails the current action is rolled back, stays in the write
         queue and an exception is raised.
@@ -84,52 +95,64 @@ class DBManager:
                     logging.debug('got write lock')
 
                     # make this a transaction
-                    db.begin_atomic()
-                    logging.debug('got atomic')
+                    with db.atomic():
+                        logging.debug('got atomic')
 
-                    if cmd == 'add':
-                        logging.debug('add')
-                        path, tags = current_item[2:]
-                        msg, _ = db.add_message(path, sync_maildir_flags=sync)
-                        logging.debug('added msg')
-                        msg.freeze()
-                        logging.debug('freeze')
-                        for tag in tags:
-                            msg.add_tag(tag, sync_maildir_flags=sync)
-                        logging.debug('added tags ')
-                        msg.thaw()
-                        logging.debug('thaw')
+                        if cmd == 'add':
+                            logging.debug('add')
+                            path, tags = current_item[2:]
+                            msg, _ = db.add(path, sync_flags=sync)
+                            logging.debug('added msg')
+                            with msg.frozen():
+                                logging.debug('freeze')
+                                for tag in tags:
+                                    msg.tags.add(tag)
+                                if sync:
+                                    msg.tags.to_maildir_flags()
+                                logging.debug('added tags ')
+                            logging.debug('thaw')
 
-                    elif cmd == 'remove':
-                        path = current_item[2]
-                        db.remove_message(path)
+                        elif cmd == 'remove':
+                            path = current_item[2]
+                            db.remove(path)
 
-                    elif cmd == 'setconfig':
-                        key = current_item[2]
-                        value = current_item[3]
-                        db.set_config(key, value)
+                        elif cmd == 'setconfig':
+                            key = current_item[2]
+                            value = current_item[3]
+                            db.config[key] = value
 
-                    else:  # tag/set/untag
-                        querystring, tags = current_item[2:]
-                        query = db.create_query(querystring)
-                        for msg in query.search_messages():
-                            msg.freeze()
-                            if cmd == 'tag':
-                                strategy = msg.add_tag
-                            if cmd == 'set':
-                                msg.remove_all_tags()
-                                strategy = msg.add_tag
-                            elif cmd == 'untag':
-                                strategy = msg.remove_tag
-                            for tag in tags:
-                                strategy(tag, sync_maildir_flags=sync)
-                            msg.thaw()
+                        else:  # tag/set/untag
+                            querystring, tags = current_item[2:]
+                            if cmd == 'toggle':
+                                presenttags = self.collect_tags(querystring)
+                                to_remove = []
+                                to_add = []
+                                for tag in tags:
+                                    if tag in presenttags:
+                                        to_remove.append(tag)
+                                    else:
+                                        to_add.append(tag)
 
-                    logging.debug('ended atomic')
-                    # end transaction and reinsert queue item on error
-                    if db.end_atomic() != notmuch.STATUS.SUCCESS:
-                        raise DatabaseError('end_atomic failed')
-                    logging.debug('ended atomic')
+                            for msg in db.messages(querystring):
+                                with msg.frozen():
+                                    if cmd == 'toggle':
+                                        for tag in to_remove:
+                                            msg.tags.discard(tag)
+                                        for tag in to_add:
+                                            msg.tags.add(tag)
+                                    else:
+                                        if cmd == 'set':
+                                            msg.tags.clear()
+
+                                        for tag in tags:
+                                            if cmd == 'tag' or cmd == 'set':
+                                                msg.tags.add(tag)
+                                            elif cmd == 'untag':
+                                                msg.tags.discard(tag)
+                                    if sync:
+                                        msg.tags.to_maildir_flags()
+
+                        logging.debug('ended atomic')
 
                     # close db
                     db.close()
@@ -203,58 +226,100 @@ class DBManager:
             raise DatabaseROError()
         self.writequeue.append(('untag', afterwards, querystring, tags))
 
+    def toggle_tags(self, querystring, tags, afterwards=None):
+        """
+        toggles tags from messages that match `querystring`.
+        This appends a toggle operation to the write queue and raises
+        :exc:`~errors.DatabaseROError` if in read only mode.
+
+        :param querystring: notmuch search string
+        :type querystring: str
+        :param tags: a list of tags to be added
+        :type tags: list of str
+        :param afterwards: callback that gets called after successful
+                           application of this tagging operation
+        :type afterwards: callable
+        :exception: :exc:`~errors.DatabaseROError`
+
+        .. note::
+            This only adds the requested operation to the write queue.
+            You need to call :meth:`DBManager.flush` to actually write out.
+        """
+        if self.ro:
+            raise DatabaseROError()
+        self.writequeue.append(('toggle', afterwards, querystring, tags))
+
     def count_messages(self, querystring):
         """returns number of messages that match `querystring`"""
-        return self.query(querystring).count_messages()
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        return db.count_messages(querystring,
+                                 exclude_tags=self.exclude_tags)
+
+    def collect_tags(self, querystring):
+        """returns tags of messages that match `querystring`"""
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        tagset = notmuch2._tags.ImmutableTagSet(
+            db.messages(querystring,
+                        exclude_tags=self.exclude_tags),
+            '_iter_p',
+            notmuch2.capi.lib.notmuch_messages_collect_tags)
+        return [t for t in tagset]
 
     def count_threads(self, querystring):
         """returns number of threads that match `querystring`"""
-        return self.query(querystring).count_threads()
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        return db.count_threads(querystring,
+                                exclude_tags=self.exclude_tags)
 
-    def _get_notmuch_thread(self, tid):
-        """returns :class:`notmuch.database.Thread` with given id"""
-        query = self.query('thread:' + tid)
-        try:
-            return next(query.search_threads())
-        except StopIteration:
-            errmsg = 'no thread with id %s exists!' % tid
-            raise NonexistantObjectError(errmsg)
+    @contextlib.contextmanager
+    def _with_notmuch_thread(self, tid):
+        """returns :class:`notmuch2.Thread` with given id"""
+        with Database(path=self.path, mode=Database.MODE.READ_ONLY) as db:
+            try:
+                yield next(db.threads('thread:' + tid))
+            except NotmuchError:
+                errmsg = 'no thread with id %s exists!' % tid
+                raise NonexistantObjectError(errmsg)
 
     def get_thread(self, tid):
         """returns :class:`Thread` with given thread id (str)"""
-        return Thread(self, self._get_notmuch_thread(tid))
+        with self._with_notmuch_thread(tid) as thread:
+            return Thread(self, thread)
 
-    def _get_notmuch_message(self, mid):
-        """returns :class:`notmuch.database.Message` with given id"""
+    @contextlib.contextmanager
+    def _with_notmuch_message(self, mid):
+        """returns :class:`notmuch2.Message` with given id"""
         mode = Database.MODE.READ_ONLY
-        db = Database(path=self.path, mode=mode)
-        try:
-            return db.find_message(mid)
-        except:
-            errmsg = 'no message with id %s exists!' % mid
-            raise NonexistantObjectError(errmsg)
+        with Database(path=self.path, mode=mode) as db:
+            try:
+                yield db.find_message(mid)
+            except:
+                errmsg = 'no message with id %s exists!' % mid
+                raise NonexistantObjectError(errmsg)
 
     def get_message(self, mid):
         """returns :class:`Message` with given message id (str)"""
-        return Message(self, self._get_notmuch_message(mid))
+        with self._with_notmuch_message(mid) as msg:
+            return Message(self, msg)
 
     def get_all_tags(self):
         """
         returns all tagsstrings used in the database
         :rtype: list of str
         """
-        db = Database(path=self.path)
-        return [t for t in db.get_all_tags()]
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        return [t for t in db.tags]
 
     def get_named_queries(self):
         """
         returns the named queries stored in the database.
         :rtype: dict (str -> str) mapping alias to full query string
         """
-        db = Database(path=self.path)
-        return {k[6:]: v for k, v in db.get_configs('query.')}
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        return {k[6:]: db.config[k] for k in db.config if
+                k.startswith('query.')}
 
-    def get_threads(self, querystring, sort='newest_first', exclude_tags=None):
+    def get_threads(self, querystring, sort='newest_first'):
         """
         asynchronously look up thread ids matching `querystring`.
 
@@ -263,38 +328,17 @@ class DBManager:
         :param sort: Sort order. one of ['oldest_first', 'newest_first',
                      'message_id', 'unsorted']
         :type query: str
-        :param exclude_tags: Tags to exclude by default unless included in the
-                             search
-        :type exclude_tags: list of str
         :returns: a pipe together with the process that asynchronously
                   writes to it.
         :rtype: (:class:`multiprocessing.Pipe`,
                 :class:`multiprocessing.Process`)
         """
         assert sort in self._sort_orders
-        q = self.query(querystring)
-        q.set_sort(self._sort_orders[sort])
-        if exclude_tags:
-            for tag in exclude_tags:
-                q.exclude_tag(tag)
-        for t in q.search_threads():
-            yield t.get_thread_id()
-
-    def query(self, querystring):
-        """
-        creates :class:`notmuch.Query` objects on demand
-
-        :param querystring: The query string to use for the lookup
-        :type query: str.
-        :returns: :class:`notmuch.Query` -- the query object.
-        """
-        mode = Database.MODE.READ_ONLY
-        db = Database(path=self.path, mode=mode)
-        q = db.create_query(querystring)
-        # add configured exclude tags
-        for tag in settings.get('exclude_tags'):
-            q.exclude_tag(tag)
-        return q
+        db = Database(path=self.path, mode=Database.MODE.READ_ONLY)
+        for t in db.threads(querystring,
+                            sort=self._sort_orders[sort],
+                            exclude_tags=self.exclude_tags):
+            yield t.threadid
 
     def add_message(self, path, tags=None, afterwards=None):
         """
